#!/usr/bin/perl
# $Id$
#####################################################################
use strict;
use POSIX qw(strftime);
use File::Temp();
use MIME::Tools;
use MIME::Parser;
use Net::IP;
use Getopt::Long;
use Pod::Usage;

#####
# sal-wrapper.pl
#
# SpamAssassin sa-learn wrapper
# based on the version from Alexandre Jousset, 2004
# (see: http://www.huschi.net/download/sa-wrapper.pl.gz)
###### 
# Copyright and License:
# see 'perldoc sal-wrapper.pl' or documentation at the end of this file
#
#####################################################################
# Version and last change date:
my $Version = 0.2.10;
my $VDate = '01/03/19';	# MM/DD/YY
#
#####################################################################
# This was written by:
#	Stefan Jakobs <projects@localside.net>
#
# Please send all comments, suggestion, bug reports, etc
#	to projects@localside.net
#
#####################################################################

#####################################################################
# configuration parameters

# specifies how much stuff will be logged to $LOGFILE.
# a value of 0 will only log the most important parts
# and a vulue of 3 will log as much as possible.
my $DEBUG = 2;
# Set to 1 and the script will log to $LOGFILE
# (must be writeable to the user which runs the script)
# Set to 0 and the script will log to STDOUT
my $LOG_TO_FILE = 1;
my $LOGFILE = '/var/log/sa-learn.log';
# set to 1 to delete the messages and the created
# directories after finishing this script.
my $CLEANUP = 1;
# specify where the script will find sa-learn and spamc
# with spamc you have to use <-L=...> instead of <--ham|--spam>
my $SPAMC    = '/usr/bin/spamc';
my $SA_LEARN = '/usr/bin/sa-learn';
# If you don't want to check if the piped messages comes
# from an intern source then you will set the following
# parameter to 0
my $LOCAL_TEST = 1;
# this script will unpack the messages to this directory
my $UNPACK_DIR = '/tmp/test';
#my $UNPACK_DIR = '/var/spool/unpack';
# specify a file which contains the IP addresses of your
# network. Generally this is the postfix networks file.
my $NETWORKS_PATH = '';
# Add IP addresses to this array, if they are not
# contained in the networks file.
# example: = qw(127.0.0.1 192.168.0.1 192.168.1.0/24)
my @ADDITIONALIP = qw(127.0.0.1);

##### variables #####################################################
my @IPinNetworks;
my $parser = new MIME::Parser;
my $log_string;
my $tmp_log = File::Temp::tmpnam(); #"$UNPACK_DIR/__log.tmp";
my $counter = 1;
my ($help, $spam, $ham, $spamham, $client);

##### options #######################################################
Getopt::Long::Configure ('no_ignore_case');
if ( @ARGV > 0) {
   GetOptions(	'help|?|h'	=> \$help,
		'debug=i'	=> \$DEBUG,
		'd+'		=> \$DEBUG,
		'f'		=> \$LOG_TO_FILE,
		's'		=> sub { $LOG_TO_FILE = 0 },
		'logfile=s'	=> sub { $LOG_TO_FILE = 1;
					 $LOGFILE = $_[1]; },
		'spam'		=> \$spam,
		'ham'		=> \$ham,
		'L=s'		=> \$spamham,
		'cleanup!'	=> \$CLEANUP,
		'localtest!'	=> \$LOCAL_TEST)
   or pod2usage(1);
}
if ($help) { pod2usage(VERBOSE => 2) }

##### check options #################################################
if ( $spamham =~ m/^(?:spam|ham|forget)$/i )
{ 
   $spamham = "-L $spamham";
   $client = $SPAMC;
} elsif ( $spam == 1 )
{ 
   $spamham = "--spam";
   $client = $SA_LEARN;
} elsif ( $ham == 1 )
{ 
   $spamham = "--ham";
   $client = $SA_LEARN;
} else
{ pod2usage(0); } 
 
##### logging #######################################################
# open the logfile for writing and share (1) it.
if ($LOG_TO_FILE) 
{
   open (DELOG, ">>$LOGFILE") or die $!;
   flock (DELOG, 1);
}

# Debug level 3
if ($DEBUG >= 3) 
{
    MIME::Tools->debugging(1);
    open(STDERR, ">>$LOGFILE");
}

##### subroutines ###################################################
# define subroutines
sub failure
{
   # Debug Level 1
   $log_string = "--f " . $_[0];
   logging("--f " . $_[0], 1);
   if ($LOG_TO_FILE) { close(DELOG); }
   die $_[0];
}

sub logging
{
   my($message, $log_level) = @_;
   # Jan  9 21:03:43
   my $now = strftime "%b %e %H:%M:%S", localtime;
   chomp($message);
   $message =~ s/\n/\n$now /g; 
   if ($LOG_TO_FILE == 1)
   {
       if ($log_level <= $DEBUG) {
          print DELOG $now ." " .$message ."\n";
       }
   } else
   {
       if ($log_level <= $DEBUG) {
          print $now ." " .$message ."\n";
       }
   } 
}
   
sub open_tmp_log
{
   open (TMP, "+>>$tmp_log") || failure("failed to open $tmp_log!");
   flock (TMP, 1);
}

sub read_close_tmp_log
{ 
   my $log_level = shift;
   my $string = shift;
   seek(TMP, 0, 0);	# go to the beginning of $tmp_log file
   my @dump = <TMP>;
   close(TMP) || failure("failed to close $tmp_log!");
   if (defined($string)) 
   {
      map {
         if ($_ =~ $string) { logging("$_", $log_level); } 
      } @dump;
   } else { map { logging("$_", $log_level); } @dump; }
   unlink($tmp_log);
}

sub cleanup
{
   if ($CLEANUP)
   {
      # Debug Level 1
      logging("--: cleanup now", 1);
      # remove the unpacked message files
      $parser->filer->purge;
      # remove the created output directory
      rmdir $parser->output_dir;
   } else { logging("--! will NOT cleanup", 0); }
   if ($LOG_TO_FILE) { close(DELOG); }
}

sub feed
{
   my $ent = shift;

   if (! $ent->parts)
   {
      # Debug Level 2
      logging("--: part number: " . $counter++, 2);
      logging($ent->body_as_string(), 2);
#         $ent->print(\*DELOG); # don't print to FH
   }

   if ($ent->head->mime_type eq 'message/rfc822') 
   {
      logging("--! message will be piped to $client.", 0);
      # Debug Level 3
      logging("--: Here is the dump of the piped message:", 3); 
#	   $ent->bodyhandle->print(\*DELOG);	# don't print to FH
      logging($ent->bodyhandle->as_string, 3); # or ->as_lines
      logging("--: end of message dump", 3);
 
      if ( $spamham != m/^-L/ )
      {
         if ($DEBUG >= 3)	# Debug Level 3 
         {
            $spamham = "-D $spamham";
         } elsif ($DEBUG >= 1)	# Debug Level 1
         {
            $spamham = "-D learn $spamham";
         }			# else: Debug Level 0
      } 

      open(SALEARN, "|$client $spamham >$tmp_log 2>&1") ||
                failure("Not able to pipe to $client: $!");
      $ent->bodyhandle->print(\*SALEARN) && 
		logging("--! message successfully piped.", 0);
      close(SALEARN);
      # workaround; $client only logs to a file
      open_tmp_log(); 
      read_close_tmp_log(1);
   }
#  elsif ($ent->head->mime_type eq 'text/html')
#  {
#     print "body of type message:\n";
#     $entity->head->print(\*STDOUT);
#     $entity->bodyhandle->print(\*STDOUT);
#  }

   my @parts = $ent->parts;
   if (@parts) {
      map { feed($_); } @parts;
   } elsif ($counter == 1) 
   { 
      logging("--! There aren't any parts "
   	."-> Nothing to feed\n--! quit", 0); 
   }
}

##### end of subroutines ############################################ 

##### CREATE IPs ####################################################
# create a array that contains all IP addresses which 
# represents our internal (local) network. 
logging("##### Starting sal-wrapper ... #####", 0);

if ($LOCAL_TEST)
{
# create Net::IP objects from the given IP address strings
# and append them to the new array
map { 
   my $ip = new Net::IP($_) || failure("failed to create IP object"); 
   @IPinNetworks = (@IPinNetworks, $ip);
} @ADDITIONALIP;

if (! $NETWORKS_PATH eq '') 
{
   open(NETS, $NETWORKS_PATH) || failure("failed to open $NETWORKS_PATH");
   my @networks = <NETS>;
   close NETS;

   # extract the IP addresses from the networks file and append
   # them to the IPinNetworks array
   map {
      if ($_ =~ /^(?:[0-9]{1,3}\.){3}[0-9]{1,3}(\/[0-9]{1,2})?/)
      {
	  # remove all whitespace characters from the IP address
	  $_ =~ s/[\s]//g;
	  my $ip = new Net::IP($_) || failure("failed to create IP object");
	  @IPinNetworks = (@IPinNetworks, $ip);
      }
   } @networks;
}

# Debug level 2
logging("--: The following IP addresses are considered as local:", 2);
map { logging($_->print(), 2); } @IPinNetworks; } 

##### end of CREATE IPs #############################################

# check if $UNPACK_DIR exists
if (! -d $UNPACK_DIR) {
  mkdir $UNPACK_DIR or failure($UNPACK_DIR ."created", 0);
}  

# don't decode the content of a message/rfc822 document 
$parser->extract_nested_messages(0);
$parser->output_under($UNPACK_DIR);

# create an object (entity) which contains the MIME message
my $entity = $parser->parse(\*STDIN) or 
	     failure("failed to parse STDIN");

# Debug level 2 
logging("--: the skeleton dump of the message:", 2);
# work around; dump_skeleton can only dump to a FH
if ($DEBUG >=2)
{
   open_tmp_log();
   $entity->dump_skeleton(\*TMP);
   read_close_tmp_log(2);
   logging("--: end of skeleton dump", 2);
}

##### LOCAL TEST ####################################################
# Check if the piped message comes from your local network
# If yes feed it to sa-learn, otherwise discard it.

if ($LOCAL_TEST)
{
#Debug level 3
logging("--: found the following \'Received:\' lines:", 3);

# parse the IP address of the sending server from every Received: 
# header line and check if it is contained in my networks array. 
for (my $i=0; $i<$entity->head->count('Received'); $i++)
{
   my $LocalMail=0;
   my $rcvd_line = $entity->head->get('Received',$i);
   # Debug level 3
   logging("$rcvd_line", 3);	

   # Received: from a.server.tld ([127.0.0.1]) by foobar
   # Received: from b.server.tld ([192.168.1.16] verified) by foobar
   # Received: from c.server.tld (c.server.tld [192.168.1.20]) by foobar
   # Received: from [192.168.1.16] (foobar) 
   if ($rcvd_line =~ /^from [a-zA-Z0-9_.-]* \([a-zA-Z0-9_.-]* ?\[((?:[0-9]{1,3}\.){3}[0-9]{1,3})\](?: verified)?\)/ || 
       $rcvd_line =~ /^from \[((?:[0-9]{1,3}\.){3}[0-9]{1,3})\] \(.*\)/)
   {
      # Debug level 2
      logging("--: extracted IP: $1", 2);
      my $ip = new Net::IP($1) || 
	       failure("creating IP object failed");
      map {
	 if ($ip->overlaps($_)==$IP_A_IN_B_OVERLAP || 
	     $ip->overlaps($_)==$IP_IDENTICAL)
	 {
	     # Debug level 1
	     logging("--: " .$_->print() 
		  ." contains " .$ip->print() 
		  ." -> seams to be from my own network\n", 1);
             $LocalMail = 1;
	 }  
      } @IPinNetworks;
      if (! $LocalMail) 
      { 
         logging("--! the message wasn't sent from my own network.", 0);
         logging("--! will exit now.", 0);
         cleanup;
         exit;
      }
   }
   else { logging("--! didn't found an IP!", 0); }
}
if (! @IPinNetworks) { cleanup; failure("No IPs in my network!"); }
}
##### end of LOCAL TEST #############################################

##### FEEDING #######################################################
# The root (0/header) part of the message was already examined
# Now look at every additional part of the piped message and 
# feed it to sa-learn if the MIME type is message/rfc822

feed($entity);

##### end of FEEDING ################################################

cleanup;

##### Documentation #################################################

=pod

=head1 NAME

B<sal-wrapper.pl> - passes attached messages to sa-learn or spamc 

=head1 SYNOPSIS

sal-wrapper.pl [-h|-?|--help] [--debug=#|-d] [-f|-s] [--logfile=<file>]
[--cleanup|--nocleanup] [--localtest|--nolocaltest] 
-L <spam|ham|forget>|--spam|--ham < message

=head1 OPTIONS

=over 8

=item B<-h|?|--help>

Prints a short help message.

=item B<-L> I<ham|spam|forget>

Script will call spamc and the message will be learned as I<ham|spam|forget>

=item B<--spam>

Script will call sa-learn and the message will be learned as spam

=item B<--ham>

Script will call sa-learn and the message will be learned as ham

=item B<--debug=>I<#>

This will set the debug output level to I<#>.
The debug level goes from 0 (no additional informations) to 3 (max logging).
Supersedes the variable $DEBUG.

=item B<-d>

Multiple flags are possible and each flag raises the debug level by one.
Increments the variable $DEBUG.

=item B<-f>

The script will print the output to the default log file or the one named
by option --logfile.
Supersedes the variable $LOG_TO_FILE.

=item B<--logfile=>I<file>

The script will print the output to I<file>.
Supersedes the variables $LOGFILE and $LOG_TO_FILE.

=item B<-s>

The script will print the output to STDOUT.
Supersedes the variable $LOG_TO_FILE.

=item B<--cleanup|--nocleanup>

The Script will/will not remove the files in the temporary directory.
Supersedes the variable $CLEANUP.

=item B<--localtest|--nolocaltest>

The Script will/will not examine the IP addresses in the Received: header
to detect if the source of the message resides in your local(own) network.
Supersedes the variable $LOCAL_TEST.

=back

=head1 DESCRIPTION

B<This program> will feed a message which is contained in the appendix
of an other message to sa-learn or spamc for training the Bayes DB.
It will unpack the inner message and pipe it to one of the SpamAssassin 
programs.

The script contains a few variables in its first section which should be
apated to the user's needs. Some, but not all, variables can be modified
by passing options to the script.
The variable I<$DEBUG> controls the talkativeness of the script. Possible
values range from 0 to 3. If I<$LOG_TO_FILE> is set to 1, then the script
will print its output to the file specified by I<$LOG_TO_FILE>. If it is 
set to 0 then all output will be written to STDOU. The
variable I<$CLEANUP> will configure if the script deletes the extracted
message parts from the temporary directory specified by I<$UNPACK_DIR>
or not. Possible values are 0 or 1. The paths to the programs spamc and
sa-learn are set by the variables I<$SPAMC> and I<$SA_LEARN>. The script
can check by examining the Received: headers if a message was send from
a specified range of IP addresses. This check is enabled if the variable
I<$LOCAL_TEST> is set to 1. 0 disables the check. The range of IP addresses
can be read from a file (e.g. Postfix' network file) or by adding them to
the array I<@ADDITIONALIP>. 

=head1 REQUIRES

Perl 5, Getopt::Long, Pod::Usage, MIME::Tools, MIME::Parser, Net::IP

=head1 COPYRIGT and LICENCE

Copyright (c) 2008-2009 Stefan Jakobs

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

=cut

